from omicimage.cli import parse_methylation_arguments
from omicimage.methylation_loader import load_scaled_cpg_methylation_data
from omicimage.sam_parser import parse_nucleotide_base_from_sam_methylation
from omicimage.utils import make_timestamped_output_directory, run_rgba_pipeline

# Main pipeline which calls the helper functions to ultimately create the pixel image.
# Parse the user input from the command line, and pass information to respective functions.
def main():
    
    cli_arguments = parse_methylation_arguments()
    methylation_data = load_scaled_cpg_methylation_data(cli_arguments.cpg_report)
    output_directory = make_timestamped_output_directory(cli_arguments.output_prefix)
    run_rgba_pipeline(parse_nucleotide_base_from_sam_methylation, cli_arguments.sam,[methylation_data, not cli_arguments.full_read, cli_arguments.mapq], cli_arguments.width, output_directory)


if __name__ == "__main__":
    main()



import numpy as np
from PIL import Image

# Takes the separate RGBA values and combines them into a single list of four elements.
def encode_pixel_rgba(base_val, meth_val, coverage_val, strand_val):
    return [base_val, meth_val, coverage_val, strand_val]

# Creates a NumPy array which respresents the image in RGBA format, the value 4 corresponds to the four color channels: Red, Green, Blue, and Alpha.
# The height and width correspond to the number of image rows and columns.
# Each "rgba" value is a list which consist of four elements [R, G, B, A]
# Returns a 3D array which respresents the full image as a Numpy array.
def create_rgba_array(image_dictionary, height, width):

    image_array = np.zeros((height, width, 4), dtype=np.uint8)

    for (row, col), rgba in image_dictionary.items():
        image_array[row, col] = rgba
    return image_array

# Convert the RGBA array into the image using the python PIL library.
def convert_rgba_array_to_image(img_array, output_path):
   
    img = Image.fromarray(img_array, mode="RGBA")
    img.save(output_path)



import os
from datetime import datetime
from omicimage.image_encoder import create_rgba_array, convert_rgba_array_to_image

# Writes the read alignment coordinates associated with each nucleotide base (pixel)
# read alignment position is a list tuples wihch contain - (row, col, chromosome, ref_pos, nucleotide_base, methylation_value, strand, base quality).
def write_read_alignment_coordinates(read_alignment_coordinates, path):
    
    with open(path, "w") as coordinates_file:
        coordinates_file.write("row\tcol\tchrom\tref_pos\tbase\tmethylation\tstrand\tbase_quality\n")
        for read_alignment in read_alignment_coordinates:
            coordinates_file.write("\t".join(map(str, read_alignment)) + "\n")

# Writes the height and width metadata on the generated image.
def write_image_metadata(path, image_width, image_height):

    with open(path, "w") as metadata_file:
        metadata_file.write(f"width: {image_width}\nheight: {image_height}\n")

# Creates an output directory and prefix the name with a user defined label and timestamp.
# The timestamp format is predefined.
def make_timestamped_output_directory(prefix):

    time_string = "Date%d_%m_%y_Time%H%M"
    timestamp = datetime.now().strftime(time_string)
    timestamp_prefix = f"{prefix}_{timestamp}"

    output_directory = os.path.join("output", timestamp_prefix)
    os.makedirs(output_directory, exist_ok=True)
    return output_directory

# Executes the full image generation pipeline using the parsed SAM data.
# Calls the specified parsing function, which returns the image dictionary, read alignment coordinates and calculated image height. 
# Construct a NumPy array which is the RGBA data for the image based on the information in the image_dictionary.
# Saves the image, read alignment position and metadata for interpreting and converting the 
# image back to the original reads in the output folder.
def run_rgba_pipeline(parsing_function, sam_path, parsing_arguments, image_width, output_directory):
    
    image_dictionary, read_alignment_coordinates, image_height = parsing_function(sam_path, *parsing_arguments, image_width=image_width)

    image_array = create_rgba_array(image_dictionary, image_height, image_width)

    convert_rgba_array_to_image(image_array, os.path.join(output_directory, "image.png"))
    write_read_alignment_coordinates(read_alignment_coordinates, os.path.join(output_directory, "readalignmentcoordinates.tsv"))
    write_image_metadata(os.path.join(output_directory, "metadata"), image_width, image_height)




import argparse

# Import argument parser in order to accurately handle user input from the command line.
# Add required and optional arguments.
# A default width is set if one is not provided.
# Default to extracting only one base from each read.
def parse_nomethylation_arguments():
    
    parser = argparse.ArgumentParser()
    parser.add_argument("sam", help="Input SAM file")
    parser.add_argument("output_prefix", help="Prefix for naming the output folder")
    parser.add_argument("--width", type=int, default=2500, help="Image width")
    parser.add_argument("--full-read", action="store_true", help="Extract all bases from each read, the default is only the first base")
    parser.add_argument("--mapq", type=int, default=None, help="Filter for mapping quality (MAPQ). If no value is provided the filtering is not applied.")
    args = parser.parse_args()

    if args.mapq is not None and args.mapq < 0:
        parser.error("--mapq must be non-negative value")

    return args

def parse_methylation_arguments():

    parser = argparse.ArgumentParser()
    parser.add_argument("sam", help="Input SAM file")
    parser.add_argument("cpg_report", help="Bisulfite Sequencing CpG report")
    parser.add_argument("output_prefix", help="Prefix for naming the output folder")
    parser.add_argument("--width", type=int, default=2500, help="Image width")
    parser.add_argument("--full-read", action="store_true", help="Extract all bases from each read, the default is only the first base")
    parser.add_argument("--mapq", type=int, default=None, help="Filter for mapping quality (MAPQ). If no value is provided the filtering is not applied.")

    args = parser.parse_args()

    if args.mapq is not None and args.mapq < 0:
        parser.error("--mapq must be non-negative value")

    return args



from omicimage.cli import parse_nomethylation_arguments
from omicimage.sam_parser import parse_nucleotide_base_from_sam
from omicimage.utils import make_timestamped_output_directory, run_rgba_pipeline

# Main pipeline which calls the helper functions to ultimately create the pixel image.
# Parse the user input from the command line, and pass information to respective functions.
def main():
    
    cli_arguments = parse_nomethylation_arguments()
    output_directory = make_timestamped_output_directory(cli_arguments.output_prefix)
    run_rgba_pipeline(parse_nucleotide_base_from_sam, cli_arguments.sam, [not cli_arguments.full_read, cli_arguments.mapq], cli_arguments.width, output_directory)

if __name__ == "__main__":
    main()



# Shared constants and functions.

MAX_BASE_QUALITY = 40

# Maps the nucleotide bases to 8-bit grayscale values.
nucleotide_BASE_TO_GRAYSCALE_MAPPER = {
    
    'A': 32,
    'T': 96,
    'C': 160,
    'G': 224,
    'N': 255
}

# Convert a Phred quality score to an 8-bit integer (0–255), 
# Clipping is used at the maximum Phred Score. 
def phred_score_to_uint8(phred_score, max_phred_score=MAX_BASE_QUALITY):

    if phred_score is None:
        return 0
    if phred_score < 0:
        phred_score = 0
    if phred_score > max_phred_score:
        phred_score = max_phred_score
    return int(round(255.0 * phred_score / max_phred_score))

# Converts a nucleotide base to its corresponding grayscale value.
# Defaults to 255 for unknown or invalid bases.
def nucleotide_base_to_grayscale(base):

    return nucleotide_BASE_TO_GRAYSCALE_MAPPER.get(base, 255)

# Checks if the read is on the reverse strand.
# Currently strand information is stored in the alpha channel: 0 for reverse, 255 for forward.
def check_strand(read):

    return 0 if read.is_reverse else 255

# Calculates the row and column in order to convert the pixel index back to its nucleotide position.
def pixel_to_base_coordinates(pixel_index, image_width):

    calculated_row = pixel_index // image_width
    calculated_column = pixel_index % image_width
    return calculated_row, calculated_column




"""SAM→RGBA image tools.

Import path: ``core``
Project distribution: ``sam2rgba``
"""
__all__ = [
    "build_rgba_methylation",
    "build_rgba_nomethylation",
    "cli",
    "image_encoder",
    "methylation_loader",
    "sam_parser",
    "shared",
    "utils",
]
__version__ = "0.1.0"



import pysam
from tqdm import tqdm

from omicimage.shared import nucleotide_base_to_grayscale, check_strand, pixel_to_base_coordinates, phred_score_to_uint8, MAX_BASE_QUALITY

# Using the pysam library to open the SAM file.
# This allows for the iteration over the sequencing reads.
def open_sam_file(path):

    return pysam.AlignmentFile(path, "r")

# Calculates the height of the image, if there is a remainder we add an additional row to accomodate it.
def calculate_image_height(image_dictionary, image_width):

    number_of_pixels = len(image_dictionary)
    image_height = number_of_pixels // image_width
    
    if number_of_pixels % image_width != 0:
        image_height += 1
    return image_height

# Inserts an RGBA pixel into the image dictionary, and appends the metadata into the read_alignment_coordinates list.
def insert_pixel(image_dictionary, read_alignment_coordinates, row, col, rgba, chromosome, ref_pos, base, methylation_value, strand, base_quality):
   
    image_dictionary[(row, col)] = rgba
    strand_symbol = '+' if strand == 255 else '-'
    read_alignment_coordinates.append((row, col, chromosome, ref_pos, base, methylation_value, strand_symbol, base_quality))

# Construct a single RGBA pixel, with base_quality and methylation_value beig set to default values.
# Clamp each channel to the range of 0–255
def construct_pixel(base, strand, methylation_value=128, base_quality=0):

    base_val = nucleotide_base_to_grayscale(base)
    strand_val = strand

    rgba = [base_val, methylation_value, base_quality, strand_val]
    
    clamped_value = []
    for channel in rgba:
        clamped_value.append(max(0, min(255, int(channel))))
    rgba = clamped_value
    return rgba

# Parses the SAM file extracting either the first or full nucleotide base/s from each aligned read.
# Extracts and stores additional metadata including the strand orientation and a placeholder methylation value.
# Encodes this information into an RGBA pixel
# Calculates the final height of the image based on the number of encoded pixels.
def parse_nucleotide_base_from_sam(sam_path, first_base_only=True, mapq_filter=None, image_width=2500):

    sam_file = open_sam_file(sam_path)
    image_dictionary = {}
    read_alignment_coordinates = []

    for read in sam_file.fetch():
        if read.is_unmapped:
            continue

        if mapq_filter is not None and read.mapping_quality < mapq_filter:
            continue

        chromosome = sam_file.get_reference_name(read.reference_id)
        strand = check_strand(read)

        # Either extract the first nucleotide from each read or extract every nucleotide base from each read within the SAM file.
        if first_base_only:
            base_index = 0
            reference_position = read.reference_start
            base_positions = [(base_index, reference_position)]
        else:
            base_positions = []
            read_length = len(read.query_sequence)
            for base_index in range(read_length):
                base_positions.append((base_index, read.reference_start + base_index))
            
            # Loop through all base in the current read.
            # Get the indexed nucleotide base in the read sequence.
            # Convert the values to a pixel and determine the position in the image.
        for base_index, ref_pos in base_positions:
            try:
                nucleotide_base = read.query_sequence[base_index].upper()
                base_quality = read.query_qualities[base_index]
                scaled_base_quality = phred_score_to_uint8(base_quality)
            except (IndexError, TypeError):
                continue  
        
            methylation_value = 128  # Default to 128 for non-methylation
            rgba = construct_pixel(nucleotide_base, strand, methylation_value, scaled_base_quality)

            image_length = len(image_dictionary)
            row, col = pixel_to_base_coordinates(image_length, image_width)

            insert_pixel(image_dictionary, read_alignment_coordinates, row, col, rgba, chromosome, ref_pos, nucleotide_base, methylation_value,strand, scaled_base_quality)

    image_height = calculate_image_height(image_dictionary, image_width)
    return image_dictionary, read_alignment_coordinates, image_height

# Parses the SAM file extracting either the first or full nucleotide base/s from each aligned read.
# Extracts and stores additional metadata including the strand orientation methylation value for the base.
# Encodes this information into an RGBA pixel
# Calculates the final height of the image based on the number of encoded pixels.

def parse_nucleotide_base_from_sam_methylation(sam_path, methylation_data, first_base_only=True, mapq_filter=None, image_width=2500):
    
    sam_file = open_sam_file(sam_path)
    image_dictionary = {}
    read_alignment_coordinates = []

    match_count = 0
    total_checked = 0

    for read in sam_file.fetch():
        if read.is_unmapped:
            continue

        if mapq_filter is not None and read.mapping_quality < mapq_filter:
            continue

        chromosome = sam_file.get_reference_name(read.reference_id)
        strand = check_strand(read)

        # Either extract the first nucleotide from each read or extract every nucleotide base from each read within the SAM file.
        if first_base_only:
            base_indices = [0]
        else:
            base_indices = range(len(read.query_sequence))

            # Loop through all base in the current read.
            # Get the indexed nucleotide base in the read sequence.
            # Convert the values to a pixel and determine the position in the image.
        for base_index in base_indices:
            try:
                nucleotide_base = read.query_sequence[base_index].upper()
                base_quality = read.query_qualities[base_index]
                scaled_base_quality = phred_score_to_uint8(base_quality)
                ref_pos = read.reference_start + base_index
            except (IndexError, TypeError):
                continue

            total_checked += 1

            # Look up methylation for this genomic position
            methylation_value = methylation_data.get((chromosome, ref_pos), 128)
            if methylation_value != 128:
                match_count += 1

            rgba = construct_pixel(nucleotide_base, strand, methylation_value, scaled_base_quality)

            image_length = len(image_dictionary)
            row, col = pixel_to_base_coordinates(image_length, image_width)

            insert_pixel(image_dictionary,read_alignment_coordinates,row, col,rgba,chromosome, ref_pos,nucleotide_base,methylation_value,strand,scaled_base_quality)

    percent_match = (match_count / total_checked * 100) if total_checked else 0
    print(f"Methylation matches: {match_count} / {total_checked} bases checked ({percent_match:.2f}%)")

    image_height = calculate_image_height(image_dictionary, image_width)
    return image_dictionary, read_alignment_coordinates, image_height




# Load the CpG report and parse the tab delimited fields.
# Extract relevant information such as the chromosome, position and methylation percentage. 
# Scale the methylation percentage (0–100) to an 8-bit value (0–255).
# Return a dictionary mapping of the chromosome and position to the scaled methylation values.
def load_scaled_cpg_methylation_data(cpg_report_path):

    scale_multiplier = 2.55
    scaled_cpg_methylation_dict = {}

    with open(cpg_report_path) as cpg_file:
        for line in cpg_file:
            if not line.strip():
                continue

            fields = line.strip().split('\t')
            if len(fields) < 4:
                continue

            chromosome = fields[0]

            try:
                cpg_position = int(fields[1])
                methylation_percent = float(fields[3])
            except ValueError:
                print("Error: Incorrect file provided.")
                print(f"Line in question: {line.strip()}")
                exit(1)  

            scaled_methylation_percentage = int(methylation_percent * scale_multiplier)
            scaled_cpg_methylation_dict[(chromosome, cpg_position)] = scaled_methylation_percentage 
    return scaled_cpg_methylation_dict
